/** @file freeGroups.h
 * Free groups with homo-, endo- and automorphisms.
 * 
 * @author Karsten Kremer
 * @version 1.0
 * @date 29.01.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */
 
#if !defined(ORIGAMI_FREEGROUPS_H_INCLUDED)
#define ORIGAMI_FREEGROUPS_H_INCLUDED

#include <cstdio>
#include <cassert>
#include <iostream>
#include <list>

using namespace std;

class FreeGroup {
	friend class F2;
	protected:	  
		list <int> rep;
		void reduce();		
	public: 
		FreeGroup() 								{}
		FreeGroup(const FreeGroup& other) 	{rep = other.rep;}
		FreeGroup(list<int> x) 					{rep = x; reduce();}
		FreeGroup(int x) 							{push_back(x);}	
		FreeGroup(int num, const int* x) 	{for (int i=0; i<num; i++) push_back(x[i]);}

		virtual ~FreeGroup() {};		
		FreeGroup copy() const  {return FreeGroup(*this);}
		
		int getLength()				const {return rep.size();}
		int back() 						const {assert(getLength()>0); return rep.back();}
		int front() 					const {assert(getLength()>0); return rep.front();}
		void pop_back() 						{assert(getLength()>0); rep.pop_back();}
		void pop_front() 						{assert(getLength()>0); rep.pop_front();}
		bool isOne() 					const {return getLength()==0;}
		void setOne() 							{rep.clear();}
		
		int countGenerator(int g) const; 
		bool isFront(int x) const;
		bool isBack(int x) const;
		bool isFrontInverse(int x) const;
		bool isBackInverse(int x) const;
		bool firstMatches (const FreeGroup& other) const;
		bool lastMatches (const FreeGroup& other) const;
		void push_back(int x); 
		void push_front(int x);
		
		FreeGroup& append(FreeGroup& other); // fast, destroys other
		FreeGroup& append_left(FreeGroup& other); // fast, destroys other
		FreeGroup& mult(const FreeGroup& other);
		FreeGroup& mult_left(const FreeGroup& other);
		FreeGroup& invert();
		FreeGroup& power(int n);
				
		FreeGroup inverse() const 			{return copy().invert();}
		
		typedef list<int>::const_iterator iterator;   // type(*it) = int
		typedef list<int>::const_reverse_iterator reverse_iterator;
		
		iterator 			begin()	const	{return rep.begin();}		
		iterator 			end()		const	{return rep.end();}
		reverse_iterator 	rbegin() const	{return rep.rbegin();}
		reverse_iterator 	rend()	const	{return rep.rend();}
				
		void operator = (const FreeGroup& other)  {rep = other.rep;}
		void operator *= (const FreeGroup& other) {mult(other);}
		void operator /= (const FreeGroup& other) {mult(other.inverse());}
//		void operator *= (int x) {push_back(x);}		
		bool operator == (const FreeGroup& other) const		{return rep == other.rep;}
		bool operator != (const FreeGroup& other) const 		{return rep != other.rep;}
		FreeGroup operator * (const FreeGroup& other) const {return copy().mult(other);}
		FreeGroup operator / (const FreeGroup& other) const {return copy().mult(other.inverse());}
		FreeGroup operator ^ (int n) const 						{return copy().power(n);}

		virtual const string generatorString(int x) const;
		virtual int inverseGenerator(int x) const  {return -x;}
		virtual bool isGenerator(int x) const  		{return (x!=0);}		
};

ostream& operator << (ostream &os, const FreeGroup& f); 

class F2 : public FreeGroup {
	private:
		char xChar;
		char yChar;
	public:
		F2() 												{setChar();}
		F2(char x, char y) 							{setChar(x,y);}
		F2(const F2& other) 							{setChar(other.xChar,other.yChar); rep = other.rep;}
		F2(int x) 										{setChar(); push_back(x);}	
		F2(const list<int> x) 						{setChar(); rep = x; reduce();}
		F2(int num, const int* x) 					{setChar(); for (int i=0; i<num; i++) push_back(x[i]);}
		
		void setChar (char x='x', char y='y') 	{xChar = x; yChar = y;}
		
		F2 copy() const 								{return F2(*this);}
		
		F2 inverse() const 									{F2 x(*this); x.invert(); 				 return x;}
		F2 operator * (const F2& other) const 		{F2 x(*this); x.mult(other); 			 return x;}
		F2 operator / (const F2& other) const 		{F2 x(*this); x.mult(other.inverse()); return x;}
		F2 operator ^ (int n) const 					{F2 x(*this); x.power(n); 				 return x;}
		void operator = (const F2& other)  			{setChar(other.xChar,other.yChar); rep = other.rep;}
		
		const string generatorString(int x) const;
		int inverseGenerator(int x) const 				{assert (isGenerator(x)); return (x+2)%4;}		
		bool isGenerator(int x) const						{return (x >=0 && x < 4);}
};

template <typename T>  class F2Map {
	private:
		T im[4];
	public:		
		F2Map() 										{T one; one.setOne(); set(one,one);}
		F2Map(const T& x, const T& y) 		{set (x,y);}
		F2Map(const F2Map<T>& other) 			{setMap (other);}
		F2Map(const T& x, const T& y, const T& xInv, const T& yInv) {set (x,y,xInv,yInv);}
		
		void set(const T& x, const T& y) 	{set(x,y,x^(-1),y^(-1));}
		void setMap(const F2Map<T>& other) 	{set(other.im[0], other.im[1], other.im[2], other.im[3]);}
		void set(const T& x, const T& y, const T& xInv, const T& yInv) {
			assert ((x*xInv == (x^0)) && (y*yInv == (y^0)));
			im[0] = x; 
			im[1] = y; 
			im[2] = xInv; 
			im[3] = yInv;
		}
		
		bool equals(const F2Map<T>& other) const {return (im[0]==other.im[0] && im[1]==other.im[1]);}
		F2Map<T> copy() const 										{return F2Map<T>(*this);}
		bool operator == (const F2Map<T>& other) const 	{return equals(other);}
		bool operator != (const F2Map<T>& other) const 	{return !equals(other);}
		void operator = (const F2Map<T>& other)  			{setMap(other);}

		const T& image (int p) const {
			assert (p >= 0 && p <= 3);
			return im[p];
		}
		
		T image (const F2& p) const {
			T result;
			result.setOne();
			for (F2::iterator it = p.begin(); it != p.end(); it++) result *= im[*it];
			return result;
		}
		
		list<T> image (const list<F2>& p) const {
			list<T> result;
			for (list<F2>::const_iterator it = p.begin(); it != p.end(); it++)
				result.push_back(image(*it));
			return result;
		}		

};

template <typename T> ostream& operator << (ostream &os, const F2Map<T>& f) {
	os << "x -> " << f.image(F2(0)) << endl;
	os << "y -> " << f.image(F2(1)) << endl;
	return os;
}

class EndF2 : public F2Map<F2> {
	public:
		EndF2()									{setOne();}
		EndF2(const F2& x, const F2& y) 	{set (x,y);}		
		EndF2(const EndF2& other) 			{setMap(other);}
		
		void setOne() 							{set(F2(0),F2(1));}
		bool isOne() const					{return equals(EndF2());}
		bool isInverse(const EndF2& other) const;
		
		EndF2& mult(const EndF2& other);
		EndF2& mult_left(const EndF2& other);
	
		EndF2 copy() const 					{return EndF2(*this);}

		void operator = (const EndF2& other)  			{setMap(other);}
		void operator *= (const EndF2& other) 			{mult(other);}
		EndF2 operator * (const EndF2& other) const 	{return copy().mult(other);}
};

class AutF2 {
	private:
		EndF2 map;
		EndF2 inv;
	public:
		AutF2() 													{setOne();}
		AutF2(const EndF2& f, const EndF2& finv)  	{set(f,finv);}
		
		void set(const EndF2& f, const EndF2& finv); 
		
		void setOne() 											{map.setOne(); inv.setOne();}
		bool isOne() const 									{return map.isOne();}
		bool equals(const AutF2& other) const			{return map.equals(other.map);}
		
		F2 image (const F2& p) const 						{return map.image(p);}
		F2 preImage (const F2& p) const 					{return inv.image(p);}
		list<F2> image (const list<F2>& p) const 		{return map.image(p);}
		list<F2> preImage (const list<F2>& p) const 	{return inv.image(p);}
		
		AutF2 inverse() const 								{return AutF2(inv,map);}
		
		AutF2& invert() 										{set(inv.copy(),map.copy()); return *this;}
		AutF2& power(int n);
		AutF2& mult(const AutF2& other); 
	
		AutF2 copy() const 									{return AutF2(this->map, this->inv);}
		
		void operator = (const AutF2& other)  			{set(other.map, other.inv);}
		bool operator == (const AutF2& other) const 	{return equals(other);}
		bool operator != (const AutF2& other) const 	{return !equals(other);}

		void operator *= (const AutF2& other) 			{mult(other);}
		AutF2 operator * (const AutF2& other) const 	{return copy().mult(other);}
		AutF2 operator / (const AutF2& other) const 	{return copy().mult(other.inverse());}
		AutF2 operator ^ (int n) const 					{return copy().power(n);}
};

ostream& operator << (ostream &os, const AutF2& f) {
	os << "x -> " << f.image(F2(0)) << endl;
	os << "y -> " << f.image(F2(1)) << endl;
	return os;
}

#endif // ORIGAMI_FREEGROUPS_H_INCLUDED

