/** @file graph.cpp
 * Adjacency graph for an origami.
 * An origami is defined by two permutations. The permutations can be loaded
 * from text-files (with file-extension .ori).
 * 
 * @author Karsten Kremer
 * @version 1.0
 * @date 29.01.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */
 
#include <cstdio>
#include <cassert>
#include <iostream>
#include <queue>
#include <stdexcept>
#include <math.h>
#include "graph.h"

using namespace std;
// ******************************* Vertex ***************************


Vertex::Vertex (const Graph& g, int idx) {
	assert(g.isVertex(idx));
	graph = &g;
	index = idx;
}

void Vertex::setIndex (int idx) {
	assert(graph->isVertex(idx));
	index = idx;
}

int  Vertex::getIndex  () const {
	return index;
}

int  Vertex::getRight  () const {
	return getNeighbour(0);
}
		
int  Vertex::getTop    () const {
	return getNeighbour(1);
}

int  Vertex::getLeft   () const {
	return getNeighbour(2);
}

int  Vertex::getBottom () const {
	return getNeighbour(3);
}

int  Vertex::getNeighbour (int dir) const {
	assert ((dir>=0) && (dir<4));
	return 	graph->getNeighbour(index,dir);
}

int  Vertex::getNeighbour (const F2& dir) const {
	return graph->getNeighbour(index,dir);
}

int Vertex::getRamificationIndex() const { // of lower left corner
	Vertex v = *this;
	int i = 0;
	do {
		v = v.getLeftV().getBottomV().getRightV().getTopV();
		i++;
	} while (v != *this && i <= graph->getSize());
	
	assert(v == *this);
	return i;		
}

ostream& operator << (ostream &os, const Vertex& v) {
	os << v.getIndex() << " " << v.getRight() << " " << v.getTop();
/*	
	os << v->getIndex() << " " << v->getRight() << " " << v->getTop() << " " << v->getLeft() << " " << v->getBottom();
	if (!v->isRoot()) os << " [" << v->getParent() << "]";
*/
//	if (!v.isRoot()) os << " [" << v.getParentDir() << "]";
	//os << " ram:" << v.getRamificationIndex();
	return os;
}



// ************************** Graph ******************+

Vertex Graph::iterator::operator * () {
	return Vertex(*g,i);
}

Graph::Graph (int n) {
	resize(n);
}

Graph::Graph(char* name) {
	int n;
	ifstream file(name);
	//assert (file); // Öffnen erfolgreich?
	if (!file) throw runtime_error("Error opening origami file.");
	file >> n;
	resize(n);

	file >> *this;
	file.close();
	
	assert(isValid());
}

Graph::Graph (int n, const Permutation& x, const Permutation& y) {
	set(n,x,y);
}

Graph::Graph (const Graph& other) {
	set(other.getSize(),other.getXperm(),other.getYperm());
}

void Graph::resize (int n) {
	size = n;
	for (int i = 0; i<=3; i++) neighbours[i].resize(size);
}

void Graph::set (int n, const Permutation& x, const Permutation& y) {
	assert (y.getSize() <= n && x.getSize() <= n);
	size = n;
	neighbours[0].set(x);
	neighbours[1].set(y);
	neighbours[2] = neighbours[0].inverse();
	neighbours[3] = neighbours[1].inverse();
}

void Graph::save(char* name) const {
	ofstream file(name);
	assert (file); // Öffnen erfolgreich?
	file << *this;
	file.close();
}

bool Graph::isVertex (int index) const {
	return ((1<=index) && (index<=size));
}

void Graph::addXedge (int a, int b) {
	neighbours[0].setImage(a,b);
	neighbours[2].setImage(b,a);
}

void Graph::addYedge (int a, int b) {
	neighbours[1].setImage(a,b);
	neighbours[3].setImage(b,a);
}

void Graph::addXYedges (int a, int b, int c) {
	addXedge(a,b);
	addYedge(a,c);
}

bool Graph::isValid () const {
	return (neighbours[0].isInverse(neighbours[2]) && neighbours[1].isInverse(neighbours[3]));
}


int Graph::getSize() const {
	return size;
} 

Permutation Graph::getCommutator() const {
	// Commutator of lower left corners (= yxy'x')
	return neighbours[1]*neighbours[0]*neighbours[3]*neighbours[2];
}

Vertex Graph::getVertex (int index) const {
	assert(isVertex(index));
	return Vertex(*this,index);
} 


int Graph::getNumberOfPoints() const {
	double R = 0.0;	
	for (int i=1; i<=getSize(); i++)
		R += 1.0/getVertex(i).getRamificationIndex();
	// each vertex with ram-index k appears k times in this sum
	assert (fabs(R-floor(R+0.5)) < 0.01);
	return (int) floor(R+0.5);	
}

int Graph::getGenus() const {
	int e = getNumberOfPoints();
	int f = getSize();
	// use g = 1 - chi/2 = 1 + (f-e)/2
	assert ((f-e) % 2 == 0);
	return 1+(f-e)/2;
}



Graph::iterator Graph::begin() const {
	return iterator(*this, 1);
}

Graph::iterator Graph::end() const {
	return iterator(*this, size+1);
}

int Graph::getNeighbour(int vertex, int dir) const {
	assert (dir >= 0 && dir <= 3);
	assert (isVertex(neighbours[dir].image(vertex)));
	return neighbours[dir].image(vertex);
}

int Graph::getNeighbour(int vertex, const F2& dir) const {
	for (F2::reverse_iterator it = dir.rbegin(); it != dir.rend(); it++)	
		vertex = getNeighbour(vertex,*it);
	return vertex;
}

void Graph::applyAutF2 (const AutF2& f, Graph& newGraph) const {
        F2 x = f.preImage(F2(0));
        F2 y = f.preImage(F2(1));

        F2Map<Permutation> m (getXperm(),getYperm());
        newGraph.set(getSize(),m.image(x),m.image(y));
}

void Graph::applyST(const F2& f, Graph& newGraph) const {
        Permutation x = getXperm();
        Permutation y = getYperm();
        Permutation buf;
        for (F2::reverse_iterator it = f.rbegin(); it != f.rend(); it++) {
                if (*it == 1)  // T
                        y = x.inverse()*y;
                else if (*it == 3)  // T^(-1)
                        y = x*y;
                else if (*it == 0)      // S
                        {x.swap(y);     x.invert();}
                else {
                        assert(*it == 2);       // S^(-1)
                        x.swap(y);      y.invert();}
        }
        newGraph.set(getSize(),x,y);
}




Graph Graph::applyST(const F2& f) const {
	Graph graph;
	applyST(f,graph);
	return graph;
}


void Graph::operator = (const Graph& other) {
	size = other.size;
	for (int i = 0; i <= 3; i++) neighbours[i] = other.neighbours[i];
}

ostream& operator << (ostream &os, const Graph& g) {
	os << g.getSize() << endl;
	for (int i=1; i<=g.getSize(); i++) {
		os << g.getVertex(i) << endl;
	}
	return os;
}

istream& operator >> (istream &is, Graph& g) {
	int idx, right, top;
	for (int i=1; i<=g.getSize(); i++) 
	{
		is >> idx >> right >> top; 
		g.addXYedges(idx,right,top);
	} 
	return is;
}

