/** @file path.cpp
 * Paths, groups of paths, and spanning trees for origamis.
 * This is used to compute the fundamental group of an origami.
 * 
 * @author Karsten Kremer
 * @version 1.1
 * @date 22.02.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */
 
#include <cstdio>
#include <cassert>
#include <iostream>
#include <list>
#include <queue>
#include "graph.h"
#include "freeGroups.h"
#include "path.h"

using namespace std;

// ************* Path *******************

Path::Path(const Vertex& startV) {
	start = startV;
	end = startV;
}

int Path::getLength() const {
	return path.getLength();
}

const Graph* Path::getGraph() const {
	return getFirst().getGraph();
}

void Path::go (int dir) {
	assert(dir >= 0 && dir < 4);
	end = getLast().getNeighbourV(dir);
	path.push_front(dir);
}
		
void Path::go (const F2& dir) {
	for (F2::reverse_iterator it = dir.rbegin(); it != dir.rend(); it++)	 go(*it);
}		
		
void Path::append (const Path& other) {
	assert(getLast()==other.getFirst());
	end = other.end;
	path.mult_left(other.path);
}

void Path::flip () {
	Vertex buf = end;
	end = start;
	start = buf;
	path.invert();
}
		
		
const F2& Path::getDir() const {
	return path;
}

int Path::getFirstDir() const {
	return path.back();
}

int Path::getLastDir() const {
	return path.front();
}
		
Vertex Path::getFirst() const {
	return start;
}

Vertex Path::getLast() const {
	return end;
}

void Path::pop_front() {
	assert(getLength()>0);
	start = getFirst().getNeighbourV(getFirstDir());
	path.pop_back();
}

void Path::pop_back() {
	assert(getLength()>0);
	end = getLast().getNeighbourV(path.inverseGenerator(getLastDir()));
	path.pop_front();
}

ostream& operator << (ostream &os, const Path& p) {
	os << p.getFirst().getIndex() << "-[" << p.getDir() << "]-" << p.getLast().getIndex();
	return os;
} 

// ********************* PathGroup

void PathGroup::add(const Path& p) {
	assert (p.getFirst() == getBase() && p.getLast() == getBase());
	generators.push_back(p.getDir());
}
		
bool PathGroup::stabilizes (const Vertex& v) const {
	for(iterator it = begin(); it != end(); it++) 
		if (v.getNeighbour(*it) != v.getIndex()) return false;  	
	return true;
}

bool PathGroup::stabilizesAll (const Graph& g) const {
	for(Graph::iterator it = g.begin(); it != g.end(); it++)
		if (!stabilizes(*it))  return false; 	  
	return true;
}

bool PathGroup::stabilizesOne (const Graph& g) const {
	for(Graph::iterator it = g.begin(); it != g.end(); it++)
		if (stabilizes(*it))  return true; 	  
	return false;
}

int PathGroup::countFixedPoints (const Graph& g) const {
	int result = 0;
	for(Graph::iterator it = g.begin(); it != g.end(); it++)
		if (stabilizes(*it)) result++;
	return result;
}

F2 PathGroup::simplify() {
	F2 result;
	bool successful;
	int count;
	do {
		successful = false;
		for (int i=0; (i<=3 && !successful); i++) {
			count = 0;
			for (iterator it=begin(); it!=end(); it++) {
				if (it->isFront(i)) count++;
				if (it->isBackInverse(i)) count++;
			}
			if (count > (int)generators.size()) {
				successful = true;
				for (list<F2>::iterator it=generators.begin(); it!=generators.end(); it++) {
					it->mult_left(F2(i).inverse());
					it->mult(F2(i));
				}				
				result *= F2(i);
				base = base.getNeighbourV(i);
			}
		}			
	} while (successful);
	return result;
}


ostream& operator << (ostream &os, const PathGroup& p) {
	os << "Base: " << p.getBase().getIndex() << endl;
	os << "Rank: " << p.getSize() << endl;
	for (PathGroup::iterator it = p.begin(); it != p.end(); it++) os << *it << endl;
	return os;
} 


// ********************  Spanning Tree


SpanningTree::SpanningTree(const Graph& graph) {
	root = graph.getVertex(1);
	parentDir = new int[graph.getSize()+1];
	findSpanningTree(1);
}

SpanningTree::SpanningTree(const Vertex& start) {
	root = start;
	parentDir = new int[root.getGraph()->getSize()+1];
	findSpanningTree(start.getIndex());
}

SpanningTree::SpanningTree(const Vertex& start, int* parentDirections) {
	root = start;
	parentDir = parentDirections;
}

SpanningTree::~SpanningTree() {
	delete parentDir;
}

const Graph* SpanningTree::getGraph()	const {
	return root.getGraph();
}

int SpanningTree::getSize() const {
	return getGraph()->getSize();
}

int SpanningTree::getParentDir(const Vertex& v)	const {
	return parentDir[v.getIndex()];
}

int SpanningTree::getParent(const Vertex& v) const {
	if (isRoot(v)) return v.getIndex();
	else	return v.getNeighbour(getParentDir(v));
}

bool SpanningTree::isParent(const Vertex& v, int dir) const {
	return (getParentDir(v.getNeighbourV(dir)) == (dir+2)%4);
}

bool SpanningTree::contains(const Vertex& v, int dir) const {
	return (getParentDir(v)==dir || isParent(v,dir));
}

Vertex SpanningTree::getParentV(const Vertex& v) const {
	if (isRoot(v)) return v;
	return v.getNeighbourV(getParentDir(v));
}

Vertex SpanningTree::getRoot() const {
	return root;
}

bool SpanningTree::isRoot(const Vertex& v) const {
	return getParentDir(v)==-1;
}

void SpanningTree::goToRoot (Path& path) const {
	while (!isRoot(path.getLast())) path.go(getParentDir(path.getLast()));
}

void SpanningTree::goToVertex (Path& path, const Vertex& target) const {
	Path other(target); 
	goToRoot(other);
	other.flip();
	goToRoot(path);
	path.append(other);
}

Path SpanningTree::getPathToRoot (const Vertex& v) const {
	Path path(v);
	goToRoot(path);
	return path;
}

Path SpanningTree::getPath (const Vertex& v, const Vertex& w)	const {
	Path path(v);
	goToVertex(path, w);
	return path;
}

PathGroup SpanningTree::getFundamentalGroup () const {
	const Graph* g = getGraph();
	PathGroup result(getRoot());
	for (Graph::iterator it = g->begin(); it != g->end(); it++) {
		for (int dir = 0; dir <= 1; dir++) {
			Path path(getRoot());
			goToVertex(path, *it);
			path.go(dir);
			goToRoot(path);	
			if (path.getLength()>0) result.add(path);
		}
	}
	return result;	
	assert (result.getSize() == g->getSize()+1);
}


PathGroup SpanningTree::getRelations() const {
	const Graph* g = getGraph();
	bool* marked = new bool[g->getSize()+1];
	for (int i = 1; i<=g->getSize(); i++)
		marked[i] = false;
	
	PathGroup result(getRoot());
	int dir[4] = {1,2,3,0}; // y x' y' x
	F2 commutator(4, dir);
	
	for (Graph::iterator it = g->begin(); it != g->end(); it++) {
		if (!marked[(*it).getIndex()]) {
			Path path = getPath(getRoot(),*it);
			do {
				path.go(commutator); // path around upper left corner
				marked[path.getLast().getIndex()] = true;
			} while (path.getLast() != *it);

			goToRoot(path);
			if (path.getLength()>0) result.add(path);
		}
	}
	delete marked;
	return result;		
}

void SpanningTree::findSpanningTree(int start) {
	const Graph* g = getGraph();
	
	assert(g->isVertex(start));
	// Breitensuche zur Bestimmung eines Spannbaums.
	
	bool* marked; // enthält die bereits besuchten Ecken
	queue<int> toDo;  // enthält die noch zu bearbeitenden Ecken
	Vertex vertex;	// aktive Ecke
	int next;	// Nachbar der aktiven Ecke
	int i;
	int countMarked;

	// Initialisierung von marked und toDo
	marked = new bool[g->getSize()+1];
	for (i = 1; i<=g->getSize(); i++) marked[i] = false;
	marked[start] = true;
	countMarked = 1;
	parentDir[start] = -1;
	toDo.push(start);
	
	while (!toDo.empty()) {
		vertex = g->getVertex(toDo.front());		
		toDo.pop();

		// Alle Nachbarn der aktiven Ecke werden markiert und in toDo gespeichert
		// Die Richtung zur Wurzel wird gespeichert.
		for (i = 0; i<4; i++) {
			next = vertex.getNeighbour(i);
			if (!marked[next]) {
				marked[next] = true;
				countMarked++;
				parentDir[next] = (i+2)%4;
				toDo.push(next);
			}
		}
	}

	assert (countMarked == g->getSize()); // zusammenhängend
	delete marked;
}

ostream& operator << (ostream &os, const SpanningTree& tree) {
	for (Graph::iterator it = tree.getGraph()->begin(); it != tree.getGraph()->end(); it++) 
		os << (*it).getIndex() << " -(" << tree.getParentDir(*it) 
									 << ")-> " << tree.getParent(*it) << endl;
	return os;
}


