/** @file permutation.h
 * Permutations of integers and their cycle-structure (called signature).
 * 
 * @author Karsten Kremer
 * @version 1.1
 * @date 22.02.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */
 
#if !defined(ORIGAMI_SYM_H_INCLUDED)
#define ORIGAMI_SYM_H_INCLUDED

#include <iostream>
#include <cstdio>
#include <cassert>
#include <set>

using namespace std;

class Permutation {
	public:
		Permutation()									{f=0; resize(1); setOne();}
		Permutation(int n)							{f=0; resize(n); setOne();}
		Permutation(const Permutation& other)	{f=0; resize(other.size); set(other);}
		~Permutation() 								{if (f != 0) delete f;}
		
		int getSize()				const {return size;}
		int image(int i)			const {return (i <= size) ? f[i] : i;}
		
		bool equals(const Permutation& other)		const;
		bool hasValidImages() 							const;
		bool isBijective()  								const;
		bool isInverse(const Permutation& other) 	const;
		bool isOne()  										const;
		int getOrder()										const;
		int getFixedPoints()								const;
		
		void setOne();
		void setImage(int x, int y);
		void resize(int n);
		void set(const Permutation& other);
		void swap(Permutation& other);
		
		Permutation copy()											const;
		Permutation mult(const Permutation& other) 			const;
		Permutation multInverse(const Permutation& other) 	const;
		Permutation power(int n) 									const;
		Permutation inverse() 										const;
		
		void invert()															{set(invArray());}
		void operator = (const Permutation& other)					{set(other);}
		void operator *= (const Permutation& other);
		Permutation operator * (const Permutation& other)const {return mult(other);}
		Permutation operator ^ (int n)							const {return power(n);}
		
		bool operator == (const Permutation& other) 		const {return equals(other);}
		bool operator != (const Permutation& other) 		const {return !equals(other);}

	private:
		int size;
		int* f;
		int* multArray(const Permutation& other)	const;
		int* invArray()	const;
		int* copyArray()	const;
		void set(int* a);
		static int max(int i, int j) 	{return (i > j) ? i : j;}
		Permutation(int n, int* a)		{size=n; f=a;}
};

ostream& operator << (ostream &os, const Permutation& p);
istream& operator >> (istream &is, Permutation& p);

class Signature {
	public:
		typedef multiset<int>::const_iterator iterator;
		Signature(const Permutation& p);
		int getSign() 			const {return getSum()-getSize()==0 ? 1 : -1;} 
		int getSum() 			const {return n;}
		int getFixedPoints() const {return count(1);}
		int getSize()			const {return orbits.size();}
		int count(int i)		const {return orbits.count(i);}
		iterator begin() 		const {return orbits.begin();}
		iterator end() 		const	{return orbits.end();}
		bool operator == (const Signature& other) const {return orbits==other.orbits;}
		bool operator != (const Signature& other) const {return orbits!=other.orbits;}
	private:
		multiset<int> orbits;
		int n;
		void calcOrbits(const Permutation& p);
};

ostream& operator << (ostream &os, const Signature& s);


// *********************** Permutation *****************************

bool Permutation::equals(const Permutation& other) const {
	for (int i=1; i<=max(size,other.size); i++) if (image(i) != other.image(i)) return false;
	return true;
}

bool Permutation::hasValidImages() const {
	for (int i=1; i<=size; i++) if (image(i)<1 || image(i)>size) return false;
	return true;
}

bool Permutation::isBijective() const {
	return (isInverse(inverse()));
}

bool Permutation::isInverse(const Permutation& other) const {
	for (int i=1; i<=max(size,other.size); i++) if (image(other.image(i)) != i) return false;
	return true;	
}


bool Permutation::isOne() const {
	for (int i=1; i<=size; i++) if (image(i) != i) return false;
	return true;
}

int Permutation::getOrder() const {
	Permutation p = *this;
	int i;
	for (i=1; !p.isOne(); i++) p *= *this;
	return i;
}

int Permutation::getFixedPoints()	const {
	int result = 0;
	for (int i=1; i<=size; i++) if (image(i) == i) result++;
	return result;	
}


void Permutation::setOne() {
	for (int i=1; i<=size; i++) f[i] = i;
}

void Permutation::setImage(int x, int y)	{
	assert(x>0 && x<=size); 
	assert(y>0 && y<=size); 
	f[x]=y;
}

void Permutation::resize(int n) {
	if (f != 0) delete f; 
	size = n; 
	f = new int[n+1]; 
}
		
void Permutation::set(int* a)	{
	if (f != 0) delete f; 
	f = a; 
} 
		
void Permutation::set (const Permutation& other) {
	if (size != other.size) resize(other.size);
	set(other.copyArray());
}	
	
void Permutation::swap(Permutation& other) {
	int* buf = f;
	f = other.f;
	other.f = buf;
	int s = size;
	size = other.size;
	other.size = s;
}
	
int* Permutation::invArray() const {
        int* g = new int[size+1];
        for (int i=1; i<=size; i++) g[i] = i;
        for (int i=1; i<=size; i++) g[image(i)] = i;
        return g;
};

		
int* Permutation::copyArray() const {
	int* g = new int[size+1];
	for (int i=1; i<=size; i++) g[i] = image(i);
	return g;
};	
		
int* Permutation::multArray(const Permutation& other) const {
	int* g = new int[max(size,other.size)+1];
	for (int i=1; i<=max(size,other.size); i++) g[i] = image(other.image(i));
	return g;
}

Permutation Permutation::copy() const {
	return Permutation(size, copyArray());
}

Permutation Permutation::mult(const Permutation& other) const {
	return Permutation(max(size,other.size), multArray(other));
}
		
Permutation Permutation::inverse() const {
	return Permutation(size, invArray());
}

Permutation Permutation::multInverse(const Permutation& other) const {
	int* g = new int[max(size,other.size)+1];
	for (int i=1; i<=max(size,other.size); i++) g[other.image(i)] = image(i);
	return Permutation(max(size,other.size), g);
}

Permutation Permutation::power(int n) const {
	if (n < 0) return inverse().power(-n);
	if (n == 1) return Permutation(*this);
	Permutation result(size);
	for (int i = 1; i <= n; i++) result *= *this;	
	return (result);
}

	
void Permutation::operator *= (const Permutation& other) {
	set(multArray(other));
	size = max(size,other.size);
}
	
ostream& operator << (ostream &os, const Permutation& p) {
	os << "[ ";
	for (int i = 1; i <= p.getSize(); i++) os << p.image(i) << " ";
	os << "]";
	return os;
}

istream& operator >> (istream &is, Permutation& p) {
	int n;
	is >> n;
	p.resize(n);
	for (int i = 1; i <= n; i++) {
		is >> n;
		p.setImage(i,n);
	}
	return is;
}


// **********************+ Signature ************************

Signature::Signature(const Permutation& p) {
	n = p.getSize();
	calcOrbits(p);
}	
	
void Signature::calcOrbits(const Permutation& p) {
	bool* marked = new bool[n+1];
	int i, k, v;
	for (i=1; i<=n; i++) marked[i] = false;
	for (i=1; i<=n; i++) {
		if (!marked[i]) {
			// calculate orbit of i
			k = 0; v = i;
			do {
				v = p.image(v);
				marked[v] = true;
				k++;
			} while (v != i && k <= n);
			assert (v == i);
			orbits.insert(k);
		}
	}
	delete marked;
}


ostream& operator << (ostream &os, const Signature& s) {
	os << "{ ";
	int val;
	Signature::iterator it = s.begin(); 
	while (it != s.end()) {
		val = *it;
		os << s.count(*it) << "x" << val << " ";
		while (it != s.end() && *it == val) it++;
	}
	os << "}";
	return os;
}

#endif // ORIGAMI_SYM_H_INCLUDED
