/** @file visualisation.cpp
 * Visualisation of Origamis (with LaTeX-output).
 * 
 * @author Karsten Kremer
 * @version 1.0
 * @date 29.01.2007
 */
 
/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */

#include <cstdio>
#include <cassert>
#include <iostream>
#include <list>
#include <stdexcept>
#include <algorithm>
#include "visualisation.h"

using namespace std;

/*		Graph* graph;
		list<Edge> edges;
		list<int> roots;
		map<Point, int> vertexMap; 
		map<int, Point> pointMap;
		int openEdges;
		*/
		

		
Point Visualisation::getMax(int layer) const {
	map<Point, int>::const_iterator it;
	Point result(layer,0,0);
	for (it = vertexMap.begin(); it != vertexMap.end() && it->first.getLayer() != layer; it++);
	for (it = vertexMap.begin(); it != vertexMap.end() && it->first.getLayer() == layer; it++) {
		result.setX(max (result.getX(), it->first.getX()));
		result.setY(max (result.getY(), it->first.getY()));
	}	
	return result;
}


		
Point Visualisation::getMin(int layer) const {
	map<Point, int>::const_iterator it;
	Point result(layer,0,0);
	for (it = vertexMap.begin(); it != vertexMap.end() && it->first.getLayer() != layer; it++);
	for (it = vertexMap.begin(); it != vertexMap.end() && it->first.getLayer() == layer; it++) {
		result.setX(min (result.getX(), it->first.getX()));
		result.setY(min (result.getY(), it->first.getY()));
	}	
	return result;
}
		
		
Visualisation::Visualisation(const Graph* g) {
	graph = g;
	clear();
}

Visualisation::Visualisation(Visualisation* other) {
	graph = other->graph;
	vertexMap = other->vertexMap; 
	pointMap = other->pointMap;
	activeLayer = other->activeLayer;
	components = other->components;
}

Visualisation::Visualisation(char* file) {
	graph = loadGraph(file);
	load (file);
}

void Visualisation::clear () {
	vertexMap.clear();
	pointMap.clear();
	activeLayer = -1;
	pMax = Point(0,0,0);
	pMin = Point(0,0,0);
	components = 1;
}

void Visualisation::addVertex(int vertex, const Point& p) {
	assert (!containsVertex(vertex));
	assert (isFittingVertex(vertex,p));
	
	vertexMap[p]=vertex;
	pointMap[vertex]=p;
	
	if (p.getLayer() == activeLayer) {
		pMax.setX(max (pMax.getX(), p.getX()));
		pMax.setY(max (pMax.getY(), p.getY()));
		pMin.setX(min (pMin.getX(), p.getX()));
		pMin.setY(min (pMin.getY(), p.getY()));
	} else {
		pMax = getMax(p.getLayer()); pMin = getMin(p.getLayer());
	}
}
		
void Visualisation::removeVertex(int vertex) {	
	assert (containsVertex(vertex));
	
	Point p = locationOf(vertex);
	
	vertexMap.erase(vertexMap.find(p));
	pointMap.erase(pointMap.find(vertex));
	
	if (pMax.getX() == p.getX() || pMax.getY() == p.getY()) pMax = getMax(countComponents()-1);		
	if (pMin.getX() == p.getX() || pMin.getY() == p.getY()) pMin = getMin(countComponents()-1);		
}
		
void Visualisation::moveVertex(int vertex, const Point& p) {	
	removeVertex(vertex);
	addVertex(vertex, p);
}

void Visualisation::moveVertex(int vertex, int dir) {	
	Point p = locationOf(vertex);
	removeVertex(vertex);
	addVertex(vertex, p.neighbour(dir));
}

void Visualisation::forceVertex(int vertex, const Point& p) {	
	if (containsVertex(vertex)) removeVertex(vertex);
	addVertex(vertex, p);
}

void Visualisation::addComponent() {
	components++;
}

void Visualisation::addComponent(int vertex) {
	components++;
	addVertex(vertex, Point(countComponents(),0,0));	
}


bool Visualisation::containsVertex (int vertex) const {
	return (pointMap.find(vertex) != pointMap.end());
}

bool Visualisation::containsPoint (const Point& p) const {
	return (vertexMap.find(p) != vertexMap.end());
}

bool Visualisation::isComplete () const {
	return (countVertices() == graph->getSize());
}

bool Visualisation::isValidPoint (const Point& p) const {
	if (containsPoint(p)) return false;
	if (p.getLayer() <= 0) return false;
	return (p.getLayer() <= components);
}

bool Visualisation::isFittingVertex (int vertex, const Point& p) const {
	if (!isValidPoint(p)) return false;
	
	const Vertex& v = graph->getVertex(vertex);
	Point NeighbourPoint;
	int nb;
	
	for (int dir=0; dir<4; dir++) {
		NeighbourPoint = p.neighbour(dir);
		if (containsPoint(NeighbourPoint)) {
			nb = vertexAt(NeighbourPoint);
			if (nb != vertex && nb != v.getNeighbour(dir)) return false;
		}
	}	
	return true;
}

bool Visualisation::isValidEdge(const Edge& e) const {
	if (!containsVertex(e.getOrigin())) return false;
	if (!e.hasValidTarget()) return false;
	if (containsVertex(e.getTarget())) return false;
	return (e.getGraph() == graph);	
}

bool Visualisation::isFittingEdge (const Edge& e) const {
	if (!isValidEdge(e)) return false;
	
	Point p = locationOf(e.getOrigin()).neighbour(e.getDirection());
	return isFittingVertex(e.getTarget(), p);
}


void Visualisation::addFittingEdge (const Edge& e) {
	assert (isFittingEdge(e));
	
	Point p = locationOf(e.getOrigin()).neighbour(e.getDirection());
	int vertex = e.getTarget();
	
	addVertex (vertex, p);
}

void Visualisation::addEdge (const Edge& e) {
	assert (isValidEdge(e));
		
	if (isFittingEdge(e))
		addFittingEdge(e);
	else {
		addComponent(e.getTarget());
	}
}

void Visualisation::addFittingEdge (int vertex, int direction) {
	addFittingEdge (Edge(graph, vertex, direction));
}

void Visualisation::addEdge (int vertex, int direction) {	
	addEdge (Edge (graph, vertex, direction));
}

int Visualisation::vertexAt (int component, int x, int y) const { // roots at (n,0,0)
	Point p(component, x, y);
	return vertexAt (p);
}

int Visualisation::vertexAt (const Point& p) const {
	map<Point, int>::const_iterator it = vertexMap.find(p);	
	assert (it != vertexMap.end()); // containsPoint(p)
	return it->second;
}

Point Visualisation::locationOf (int vertex) const {
	map<int, Point>::const_iterator it = pointMap.find(vertex);	
	assert (it != pointMap.end()); // containsVertex(vertex)
	return it->second;
}

int Visualisation::countVertices() const {
	return (pointMap.size());
}

int Visualisation::countComponents() const {
	return (components);
}

int Visualisation::getDiameter(int component) const {
	Point p = getMax(component) - getMin(component);
	return max (p.getX(), p.getY());
}

int Visualisation::getLastDiameter() const {
	Point p = pMax - pMin;
	return max (p.getX(), p.getY());
}


bool Visualisation::operator < (const Visualisation& other) const {
	if (countVertices() != other.countVertices()) 
		return countVertices() < other.countVertices();
	if (countComponents() != other.countComponents()) 
		return countComponents() > other.countComponents();
	return getLastDiameter() > other.getLastDiameter();
}

void Visualisation::save(const char* name) const {
	map<Point, int>::const_iterator it;
	int layer = 1;
	int dir;
	Point p;
	Vertex v;
	ofstream os(name);
	assert (os); // Öffnen erfolgreich?

	os << "% Origami of size " << graph->getSize() << "\n";
	os << "\\begin{xy}\n<0.8cm,0cm>:\n";
	for (it = vertexMap.begin(); it != vertexMap.end(); it++) {
		p = it->first; 
		v = graph->getVertex(it->second);
		if (p.getLayer() != layer) {
			layer = p.getLayer();
			os << "\\end{xy}\n";
			os << "\\begin{xy}\n<0.8cm,0cm>:\n";
		}
		os << "(" << p.getX() << "," << p.getY() << ")";
		os << "*{\\OriSquare{" << it->second << "}"; 
		for (dir = 0; dir <4; dir++) {
			if (!containsPoint(p.neighbour(dir)))
				os << "{" << v.getNeighbour(dir) << "}";
			else os << "{}";
		}	
		os << "};\n";
	}	
	os << "\\end{xy}\n";
	os.close();
}

void Visualisation::load(const char* name) {
	string s;
	char c;
	int x,y,n;
	int layer = 0;
	ifstream is(name);
	//assert (is); // Öffnen erfolgreich?
	if (!is) throw runtime_error("Error opening visualisation file.");

	clear();
	components = 0;
	is.ignore(18); // % Origami of size 
	is >> n;
	while (!is.eof()) {
		is >> s; // \begin{xy]
		is >> s; // <0.8cm,0cm>:
		is >> c; // \ oder (
		layer++;
		addComponent();
		while (c == '(') {
			is >> x >> c >> y >> c;
			is.ignore(13);
			is >> n;
			addVertex(n, Point(layer,x,y));
			is >> s;
			is >> c;
		}
		is >> s; // end{xy}
	}
	is.close();
}


Graph* Visualisation::loadGraph(const char* name) {
	Graph* g;
	string s;
	char c;
	int x,y,n;
	ifstream is(name);
	//assert (is); // Öffnen erfolgreich?
	if (!is) throw runtime_error("Error opening visualisation file.");
	is.ignore(18); // % Origami of size 
	is >> n;
	g = new Graph(n);
	
	while (!is.eof()) {
		is >> s; // \begin{xy]
		is >> s; // <0.8cm,0cm>:
		is >> c; // \ oder (
		while (c == '(') {
			is >> x >> c >> y >> c;
			is.ignore(13); is >> n;
			is.ignore(2);  is >> x;
			is.ignore(2);  is >> y;
			g->addXYedges(n,x,y);
			is >> s;
			is >> c;
		}
		is >> s; // end{xy}
	}
	is.close();
	assert(g->isValid());
	return g;
}

void Visualisation::expand(int vertex, int radius) {
	if (radius == 0) radius = graph->getSize();
	
	const Vertex& v = graph->getVertex(vertex);
	list<Edge>* openEdges = Edge::neighbourEdges(v);
	list<Edge>* newEdges;
	list<Edge>::iterator it = openEdges->begin();
	
	bool* visited = new bool[graph->getSize()+1];
	for (int i=1; i<=graph->getSize(); i++) visited[i]=false;
	visited[vertex] = true;

	Point pStart = locationOf(vertex);
	Point p;
	Edge edge = openEdges->front(); 	
	
	// ***** Loop *****
	
	while (!isComplete() && !openEdges->empty()) {
		edge = openEdges->front(); 
		openEdges->pop_front();
				
		if (isFittingEdge(edge)) addEdge(edge);
		
		p = locationOf(edge.getOrigin()).neighbour(edge.getDirection());
				
		if (containsVertex(edge.getTarget()))
			if (locationOf(edge.getTarget())==p 
					&& (p-pStart).maxNorm() <= radius
					&& !visited[edge.getTarget()]) {
				
				visited[edge.getTarget()] = true;
				newEdges = edge.getTargetEdges();
				openEdges->splice(openEdges->end(), *newEdges);
				delete newEdges;
			} 
	}
	
	delete visited;
	delete openEdges;
}
	
	
/*	// ***** Initialisation *****
	if (minVert == 0) minVert = graph->getSize();
	if (maxDia == 0) maxDia = graph->getSize();
	Vertex* v = graph->getRoot();
	int bestDiameter = 0;
	Visualisation* vis = new Visualisation(v);	
	Visualisation* best = new Visualisation(v);
	list<Edge>* openEdges = Edge::neighbourEdges(v);
	list<Edge>* newEdges;
	list<Edge>::iterator it = openEdges->begin();
	
	int vertex;
	Edge edge = *it;
			
	// ***** Loop *****
	do {
		if (best->isComplete() && vis->getLastDiameter() > bestDiameter) it = openEdges->end();
		
		if (!vis->isComplete())
			while (it != openEdges->end() && !vis->isFittingEdge(*it)) it++;	
		
		if (!vis->isComplete() && it != openEdges->end()) {
			edge = *it;
			vis->addEdge(edge);
			newEdges = edge.getTargetEdges();
			openEdges->splice(openEdges->end(), *newEdges);
			delete newEdges;
			if (*best < *vis) {
				*best = *vis;
				bestDiameter = vis->getLastDiameter();
//				cout << best;
				cout << "vertices: " << best->countVertices() << " of " << graph->getSize();
				cout << "  components: " << best->countComponents();
				cout << "  diameter: " << bestDiameter << endl;
				if (best->countVertices() >= minVert && bestDiameter <= maxDia) {
					delete openEdges;
					delete vis;
					cout << endl;
					return best;	
				}					
			}
		} else if (vis->countVertices() > 1) {
			edge = vis->getLastEdge();
			vertex = edge.getTarget();
			vis->removeLastEdge();
			while (openEdges->back().getOrigin() == vertex) openEdges->pop_back();
			it = find(openEdges->begin(),openEdges->end(),edge);
			it++;
		} else {
			delete openEdges;
			delete vis;
			cout << endl;
			return best;
		}	
	} while (true);

}*/

