/** @file visualisation.h
 * Visualisation of Origamis (with LaTeX-output).
 * 
 * @author Karsten Kremer
 * @version 1.0
 * @date 29.01.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */
 
#if !defined(ORIGAMI_VISUALISATION_H_INCLUDED)
#define ORIGAMI_VISUALISATION_H_INCLUDED

#include <cstdio>
#include <cassert>
#include <iostream>
#include <list>
#include <map>
#include "graph.h"

using namespace std;

class Graph;


		class Edge {
			private: 
				const Graph* graph;
				int vertex;
				int direction;
			public:
				Edge (const Graph* g, int v, int dir = -1) {
					assert (g->isVertex(v));
					graph = g;
					vertex = v;
					direction = dir;					
				}
				Edge (const Vertex& v, int dir = -1) {
					graph = v.getGraph();
					vertex = v.getIndex();
					direction = dir;					
				}
				bool hasValidTarget() const {return (direction >= 0 && direction < 4);}
				const Graph* getGraph() const {return graph;}
				int getDirection() const {return direction;}
				int getOrigin() const {return vertex;}
				Vertex getOriginV() const  {return graph->getVertex(vertex);}
				int getTarget() const  {
					assert (hasValidTarget()); 
					return getOriginV().getNeighbour(direction);
				}
				Vertex getTargetV() const {return graph->getVertex(getTarget());}
				static list<Edge>* neighbourEdges (const Vertex& vertex)  {					
					list<Edge>* Edges = new list<Edge>;
					for (int dir=0; dir<4; dir++) 
						Edges->push_back(Edge(vertex, dir));	
					return Edges;					
				}
				list<Edge>* getTargetEdges ()  const {
					const Vertex& vertex = getTargetV();
					list<Edge>* Edges = new list<Edge>;
					for (int dir=0; dir<4; dir++) 
						if ((dir - direction + 2) % 4 != 0) Edges->push_back(Edge(vertex, dir));	
					return Edges;					
				}
				bool operator == (const Edge& other) const {
					return (graph==other.graph && vertex == other.vertex && direction == other.direction);
				}
		};


		class Point {
			friend ostream& operator << (ostream &os, const Point& p) {
				os << "(" << p.layer << "," << p.x << "," << p.y << ")";
				return os;
			}
			private: 
				int layer;
				int x;
				int y;
			public:
				Point (int a, int b, int c) {set(a,b,c);}
				Point (const Point& other) {set(other);}
				Point () {set(1,0,0);}				
				void set (int a, int b, int c) {layer = a; x=b; y=c;}
				void set (const Point& other) {set(other.layer,other.x,other.y);}
				void setLayer (int l) {layer = l;}				
				void setX (int n) {x = n;}				
				void setY (int n) {y = n;}				
				int getLayer () const {return layer;}				
				int getX () const {return x;}				
				int getY () const {return y;}				
				int maxNorm () const {return (max(abs(x),abs(y)));}
				Point neighbour (int dir) const {
					Point result(*this);
					if (dir%2==0) result.x += (dir==0)?1:-1;
					if (dir%2==1) result.y += (dir==1)?1:-1;
					return result;
				}
				bool operator < (const Point& other) const  {
					if (layer != other.layer) return (layer < other.layer);
					if (x != other.x) return (x < other.x);
					return (y < other.y);
				}
				bool operator == (const Point& other) const  {
					return (layer == other.layer && x == other.x && y == other.y);
				}
				Point operator + (const Point& other) const {
					return Point(layer+other.layer, x+other.x, y+other.y);
				}
				Point operator - (const Point& other) const {
					return Point(layer-other.layer, x-other.x, y-other.y);
				}				
		};

class Visualisation {
	private:
		const Graph* graph;
		map<Point, int> vertexMap; 
		map<int, Point> pointMap;
		int activeLayer;
		int components;
		Point pMax;
		Point pMin;
		Point getMax(int layer) const;
		Point getMin(int layer) const;
	public:
		Visualisation(const Graph* g);
		Visualisation(Visualisation* other);
		Visualisation(char* file);
		const Graph* getGraph() const {return graph;};
		void clear();
		void addVertex(int vertex, const Point& p);
		void removeVertex (int vertex);
		void moveVertex(int vertex, const Point& p);
		void moveVertex(int vertex, int dir);
		void forceVertex(int vertex, const Point& p);
		void addComponent();
		void addComponent(int vertex);
		bool containsVertex (int vertex) const ;
		bool containsPoint (const Point& p) const ;
		bool isComplete () const ;
		bool isRoot (int vertex) const ;
		bool isValidPoint (const Point& p) const;
		bool isFittingVertex (int vertex, const Point& p) const;
		bool isValidEdge(const Edge& e) const;
		bool isFittingEdge (const Edge& e) const;
		void addFittingEdge (const Edge& e);
		void addFittingEdge (int vertex, int direction);
		void addEdge (const Edge& e); 
		void addEdge (int vertex, int direction);
		void expand (int vertex, int radius=0);
		int vertexAt (int component, int x, int y) const ; // roots at (n,0,0)
		int vertexAt (const Point& p) const ;
		Point locationOf (int vertex) const ;
		int countVertices() const ;
		int countComponents() const ;
		int getDiameter(int component) const ;
		int getLastDiameter() const ;
		void load(const char* name);
		static Graph* loadGraph(const char* name);
		void save(const char* name) const;
		bool operator < (const Visualisation& other) const ;
};


/*ostream& operator << (ostream &os, const list<Edge> &edges) {
	list<Edge>::const_iterator it;
	for (it = edges.begin(); it != edges.end(); it++)
		os << it->getTarget() << " ";
	return os;
};*/



#endif // ORIGAMI_VISUALISATION_H_INCLUDED
