/** @file visualize.x.cpp
 * Executable visualising an origami (with LaTeX-output).
 * Needs the file "origami.tex" in the current directory. 
 * 
 * Usage: call the programm with an *.ori-file as parameter.
 * Then enter h for a listing of possible commands.
 * A good start will be to enter for example x1 6626,
 * which will walk from sqare number 1 two steps to the right (66),
 * then one step down (2), and another step to the right.
 * Open the created file "origami.dvi" to see the result.
 *
 * @author Karsten Kremer
 * @version 1.0
 * @date 29.01.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */
 
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cassert>
#include <list>
#include "freeGroups.cpp"
#include "graph.cpp"
#include "path.cpp"
#include "visualisation.cpp"

using namespace std;

int cToDir (char c) {
	switch (c) {
		case '2': return 3;
		case '4': return 2;
		case '6': return 0;
		case '8': return 1;
		default: return -1;
	}
}

void showVis(Visualisation* vis) {
	vis->save("testorigami.vis");
	system("latex -interaction=nonstopmode origami.tex");
}


void addNeighbours(Visualisation* vis, int vertex, string directions) {
	int dir;
	Vertex v = vis->getGraph()->getVertex(vertex);
	Point p = vis->locationOf(vertex);
	
	for (unsigned int i=0; i<directions.length(); i++) {
		dir = cToDir( directions[i]);
		if (dir == -1) {cout << "Error: Invalid direction." << directions[i] << endl; return;}
		else {
			v = v.getNeighbourV(dir);
			vertex = v.getIndex();
			p = p.neighbour(dir);
			if (!vis->containsPoint(p)) {
				if (!vis->isValidPoint(p))
					{cout << "Error: Invalid point for vertex " << vertex << endl;	return;}
				if (!vis->isFittingVertex(vertex,p)) 
					{cout << "Error: Vertex " << vertex << " doesn't fit." << endl;	return;}
				vis->forceVertex(vertex, p);
			}
		}
	}
	return;
}

Point readPoint(istream& is) {
	int x,y;
	char c;
	is >> c >> x; 
	if (c != '@') cout << "Warning: Syntax error in given Point. Unexpected character " << c << endl; 
	is >> c >> y;
	if (c != ',') cout << "Warning: Syntax error in given Point. Unexpected character " << c << endl; 
	return Point(1,x,y);
}

Point getRelativePoint(Point point, string directions) {
	int dir;
	Point p = point;
	for (unsigned int i=0; i<directions.length(); i++) {
		dir = cToDir(directions[i]);
		if (dir == -1) {cout << "Error: Invalid direction." << directions[i] << endl; return p;}
		else p = p.neighbour(dir);
	}
	return p;
}


void changeVis(Visualisation *vis, char cmd, int vertex, istream& is) {
	Point p;
	char c;
	string s;
	int n;
	
	const Graph* graph = vis->getGraph();
	

	if (!graph->isVertex(vertex)) {cout << "Error: No valid vertex entered." << endl; return;}
	
	if (cmd == '+' || cmd == 'm' || cmd == 'M') {
		while (is.peek()==' ') is.ignore(1);
		c = is.peek();
		if (c=='@') p = readPoint(is);
		else {
			is >> s;
			if (!vis->containsVertex(vertex)) {cout << "Error: Vertex not visible." << endl;	return;}	
			p = getRelativePoint(vis->locationOf(vertex), s);
		}
	}
	
	cout << "Vertex: " << vertex << " - Command: " << cmd << " - Point: " << p << " - String: >" << s << "<" << endl;
	
	if (cmd == '-') {
		if (!vis->containsVertex(vertex)) {cout << "Error: Vertex not visible." << endl;	return;}
		vis->removeVertex(vertex);		
	} else if (cmd == '+') {
		if (vis->containsVertex(vertex)) {cout << "Error: Vertex already visible." << endl;	return;}
		if (!vis->isValidPoint(p)) {cout << "Error: Invalid point." << endl;	return;}
		if (!vis->isFittingVertex(vertex,p)) {cout << "Error: Vertex doesn't fit." << endl;	return;}
		vis->addVertex(vertex, p);
	} else if (cmd == 'm') {
		if (!vis->containsVertex(vertex)) {cout << "Error: Vertex not visible." << endl;	return;}
		if (!vis->isValidPoint(p)) {cout << "Error: Invalid point." << endl;	return;}
		if (!vis->isFittingVertex(vertex,p)) {cout << "Error: Vertex doesn't fit." << endl;	return;}
		vis->moveVertex(vertex, p);
	} else if (cmd == 'M') {
		is >> s;
		cout << "Error: Command M not yet implemented." << endl;	return;
	} else if (cmd == 'x') {
		is >> s;
		if (!vis->containsVertex(vertex)) {cout << "Error: Vertex not visible." << endl;	return;}
		addNeighbours(vis,vertex,s);
	} else if (cmd == 'X') {
		is >> n;
		if (!vis->containsVertex(vertex)) {cout << "Error: Vertex not visible." << endl;	return;}
		vis->expand(vertex, n);
	} else {
		cout << "Error: Unknown command." << endl; return;
	}
	
	showVis(vis);
	cout << endl;
	cout << "Vertex: " << vertex << " - Command: " << cmd << " - Point: " << p << " - String: >" << s << "<";
	cout << endl << endl;

}


void mainMenu(Visualisation* vis, istream& is) {
	char command;
	int vertex;
	Point p;
	string s;
	showVis(vis);
//	system("kdvi origami.dvi");
	cout << endl;
	cout << "Your origami was saved to origami.dvi." << endl;
	cout << "Change of Visualisation: (type h for help, q to quit)" << endl;
	cout << "(" << vis->countVertices() << "/" << vis->getGraph()->getSize() << ") >";
	is >> command;
	while (command != 'q') {
		if (command != 'h') while (is.peek()==' ') is.ignore(1);
		switch (command) {
			case 'h':
				cout << "Syntax for commands:" << endl;
				cout << "h           help" << endl;
				cout << "q           quit" << endl;
				cout << "l [file]    load a *.vis file" << endl;
				cout << "s [file]    save a *.vis file" << endl;
				cout << "f[V]        find a vertex" << endl;
				cout << "+[V] @[Pt]  insert Vertex V at Point Pt" << endl;
				cout << "-[V]        delete Vertex V"  << endl;
				cout << "m[V] [dir]  move Vertex V in the given direction"  << endl;
				cout << "m[V] @[Pt]  move Vertex V to Point Pt"  << endl;
//				cout << "M[V] [dir]  move the connected component of V in the given direction"  << endl;
//				cout << "M[V] @[Pt]  move the connected component of V in the given direction"  << endl;
				cout << "x[V] [dir]  start at Vertex V and add the next Vertex in the given direction"  
					  << endl;
				cout << "X[V] [rad]  start at Vertex V and add possible new vertices up to a given radius"
					  << endl;
			
				cout << endl;
				cout << "[dir] is a (sequence of) the direction(s) 2 (down), 4 (left), 6 (right) or 8 (up)"
					  << endl;
				cout << "[Pt]  is a point of the form x,y (usually, Vertex 1 will start at coordinates 	0,0)" 
						<< endl;
				break;
			case 'l':
				is >> s;
				vis->load(s.c_str());
				showVis(vis);
				cout << "Loaded " << s << endl;
				break;
			case 's':	
				is >> s;
				vis->save(s.c_str());
				cout << "Saved " << s << endl;
				break;
			case 'f':
				is >> vertex;				
				if (vis->containsVertex(vertex)) {
					p = vis->locationOf(vertex);
					cout << "Position of vertex " << vertex << ": (" << p.getX() << "," << p.getY() << ")"
						  << endl;
				} else cout << "Vertex not visible." << endl;
				break;
			default:
				is >> vertex;
				changeVis (vis, command, vertex, is);
		}
		cout.flush(); is.clear();
		cout << "(" << vis->countVertices() << "/" << vis->getGraph()->getSize() << ") >";
		is >> command;
	}							
		
}


bool hasExtension(char* s, char* ext) {
	int n = strlen(s);
	int nExt = strlen(ext);
	if (nExt > n) return false;
	return (strcmp(&s[n-nExt], ext) == 0); 
}


int main (int argc, char* argv[]) {
	Visualisation* vis;
	string s;
	
	if (argc > 1) {
		if (hasExtension(argv[1],".ori")) {
			vis = new Visualisation(new Graph(argv[1]));
			if (argc > 2) {
				if (hasExtension(argv[2],".vis")) 
					vis->load(argv[2]);
			} else vis->addVertex (1, Point(1,0,0));
//		}	else if (hasExtension(argv[1],".vis")) {
//			vis = new Visualisation(argv[1]);
		}	else {
			cout << "Enter name of Origami-File as Parameter (*.ori)." << endl;
			cout << "You may add a visualisation to load (*.vis)." << endl;
				return 0;
			}
	}
	else {
		cout << "Enter name of Origami-File as Parameter (*.ori)." << endl;
		cout << "You may add a visualisation to load (*.vis)." << endl;
		return 0;
	}
	
	mainMenu(vis, cin);
	
	delete vis->getGraph();
	delete vis;
	return(0);
}

